from OSC import OSCServer
# FIXME: don't print  tracebacks in final implementation?
OSCServer.print_tracebacks = True

from ..driver import EventDriver
from ..events import PointDownEvent, PointMoveEvent, PointUpEvent
from ..touch_objects import TouchPoint
from ..screen import pixel_coords


class TuioDriver(EventDriver):
    tuio_address = 'localhost', 3333

    def __init__(self, root_area=None):
        super(TuioDriver, self).__init__(root_area)

        # OSC server that listens to incoming TUIO events
        self.server = OSCServer(self.tuio_address)
        self.server.addDefaultHandlers()
        self.server.addMsgHandler('/tuio/2Dobj', self.receive)
        self.server.addMsgHandler('/tuio/2Dcur', self.receive)
        self.server.addMsgHandler('/tuio/2Dblb', self.receive)

        # List of alive session id's
        self.alive = set()

        # List of session id's of points that have generated a 'point_down'
        # event
        self.down = set()

        # Map of session id to touch point
        self.points = {}

    def receive(self, addr, tags, data, source):
        surface = addr[8:]
        #self.debug('Received message <surface=%s tags="%s" '
        #           'data=%s source=%s>' % (surface, tags, data, source))
        msg_type = data[0]

        # FIXME: obj/blb events are ignored (for now)
        if surface != 'cur':
            return

        if msg_type == 'alive':
            alive = set(data[1:])
            released = self.alive - alive
            self.alive = alive

            if released:
                self.debug('Released %s.' % ', '.join(map(str, released)))
                self.down -= released

            for sid in released:
                point = self.points[sid]
                del self.points[sid]
                self.delegate_event(PointUpEvent(point))
        elif msg_type == 'set':
            sid, x, y = data[1:4]

            if sid not in self.alive:
                raise ValueError('Point with sid %d is not alive.' % sid)

            # Translate to pixel coordinates
            px, py = pixel_coords(x, y)

            # Check if 'point_down' has already been triggered. If so, trigger
            # a 'point_move' event instead
            if sid in self.down:
                self.debug('Moved %d to (%.4f, %.4f), in pixels: (%d, %d).'
                           % (sid, x, y, px, py))
                point = self.points[sid]
                point.set_position(px, py)
                self.delegate_event(PointMoveEvent(point))
            else:
                self.debug('Down %d at (%.4f, %.4f), in pixels: (%d, %d).'
                           % (sid, x, y, px, py))
                self.down.add(sid)
                self.points[sid] = point = TouchPoint(px, py)
                self.delegate_event(PointDownEvent(point))

    def start_loop(self):
        self.info('Starting OSC server')
        self.server.serve_forever()

    def stop(self):
        self.info('Stopping OSC server')
        self.server.close()

    def run(self):
        self.server.handle_request()
